<?php
# interface to popular dbs
# NopDB extends NopSQL <- (mysqli / PDO)

/**
 * 2017-02-07 changed NopDB_.. constants to strings
 * 2017-02-08 added mg_dlc_file(), mg_dlc_load(),  mg_dlc_save()
 */

define ('NopDB_wp', "WordPress"); // WordPress (blog, cms)
define ('NopDB_mg', "Magento"); // Magento (shop)
define ('NopDB_dr', "Drupal"); // Drupal (cms)
define ('NopDB_jo', "Joomla"); // joomla (cms)
define ('NopDB_t3', "TYPO3"); // typo3 (cms)
define ('NopDB_ps', "PrestaShop"); // prestashop (shop)
define ('NopDB_zc', "zen-cart"); // zen-cart (shop)
define ('NopDB_oc', "opencart"); // OpenCart (shop)
define ('NopDB_m2', "Magento-2"); // Magento-2 (shop)

if(!defined('CR')) define('CR', "\r");
if(!defined('LF')) define('LF', "\n");
if(!defined('CRLF')) define('CRLF', CR.LF);
if(!defined('crlf')) define('crlf', CRLF);

/* blank method-template
    case NopDB_wp: break;
    case NopDB_mg: break;
    case NopDB_dr: break;
    case NopDB_jo: break;
    case NopDB_t3: break;
    case NopDB_ps: break;
    case NopDB_zc: break;
    case NopDB_oc: break;
    case NopDB_m2: break;
*/

require_once('nopro_sqlX.php');
class NopDB extends NopSQL {

  public
    $pop = 0,
    $version = null,
    $base = null,
    $conf = false,
    $salt = null,
    $admin = null
    ;

  public function info($verb = false) {
    // return array with basic settings
    $keys = array(
		  'pop',
		  'base',
		  'conf',
		  'version',
		  'dbhost',
		  'dbname',
		  'dbuser',
		  'dbpass',
		  'prefix',
		  'salt',
		  'admin',
		  );
    $info = array();
    foreach($keys as $k) {
      if($verb || !empty($this->$k))
	$info[$k] = $this->$k;
    }
    return $info;
  }

  public function conf($type) {
    // return relative path to config-file
    switch($type) {
    case NopDB_wp: return 'wp-config.php';
    case NopDB_mg: return 'app/etc/local.xml';
    case NopDB_dr: return 'sites/default/settings.php';
    case NopDB_jo: return 'configuration.php';
    case NopDB_t3: return 'typo3conf/LocalConfiguration.php';
    case NopDB_ps: return 'config/settings.inc.php';
    case NopDB_zc: return 'includes/configure.php';
    case NopDB_oc: return 'upload/config.php';
    case NopDB_m2: return 'app/etc/env.php';
    default: return false;
    }
  }
  
  function __destruct() { if($this->pop) parent::__destruct(); }
  function __construct($type, $basedir, $dbdrvr = "mysqli") {
    $this->dbdrvr = $dbdrvr;
    $basedir = rtrim($basedir, "/");
    $conf = $this->conf($type);
    if($conf === false || !file_exists("$basedir/$conf")) die("configuration '$basedir/$conf' not found\n");
    $this->pop = $type; // assign this only after config-file was found
    $this->base = $basedir;
    $this->conf = "$basedir/$conf";
    echo "init(".$this->conf."): ".LF;
    $err = error_reporting(E_ALL & ~E_WARNING & ~E_NOTICE);
    switch($this->pop) {
    case NopDB_wp:
      include($this->conf);
      $this->dbhost = DB_HOST;
      $this->dbname = DB_NAME;
      $this->dbuser = DB_USER;
      $this->dbpass = DB_PASSWORD;
      $this->prefix = $table_prefix;
      $this->salt   = '';
      
      if(empty($wp_version))
	include(ABSPATH . WPINC . '/version.php');
      $this->version = $wp_version;
	
      break;

      
    case NopDB_mg:
      $xml = new SimpleXMLElement($this->conf, 0, true);
      #print_r($xml);
      #print_r($xml->global->resources);
	
      $this->dbhost = (string) $xml->global->resources->default_setup->connection->host;
      $this->dbname = (string) $xml->global->resources->default_setup->connection->dbname;
      $this->dbuser = (string) $xml->global->resources->default_setup->connection->username;
      $this->dbpass = (string) $xml->global->resources->default_setup->connection->password;
      $this->prefix = (string) $xml->global->resources->db->table_prefix;
      $this->salt   = (string) $xml->global->crypt->key;
      $this->admin  = (string) $xml->admin->routers->adminhtml->args->frontName;
      $xml = null;
      break;
      
    case NopDB_dr:
      include($this->conf);
      $db = $databases['default']['default'];
      $this->dbhost = $db['host'];
      $this->dbname = $db['database'];
      $this->dbuser = $db['username'];
      $this->dbpass = $db['password'];
      $this->prefix = $db['prefix'];
      $this->salt   = $drupal_hash_salt;
      break;
      
    case NopDB_jo:
      include($this->conf);
      $j = new Jconfig();
      $this->dbhost = $j->host;
      $this->dbname = $j->db;
      $this->dbuser = $j->user;
      $this->dbpass = $j->password;
      $this->prefix = $j->dbprefix;
      $this->salt   = '';
      $j = null;
      break;

    case NopDB_t3:
      function t3r($f) { return include($f); }
      $t3 = t3r($this->conf);
      $this->dbhost = $t3['DB']['host'];
      $this->dbname = $t3['DB']['database'];
      $this->dbuser = $t3['DB']['username'];
      $this->dbpass = $t3['DB']['password'];
      $this->prefix = '';
      $this->salt   = '';
      break;

    case NopDB_ps:
      include($this->conf);
      $this->dbhost = _DB_SERVER_;
      $this->dbname = _DB_NAME_;
      $this->dbuser = _DB_USER_;
      $this->dbpass = _DB_PASSWD_;
      $this->prefix = _DB_PREFIX_;
      $this->salt   = _COOKIE_KEY_;
      break;

    case NopDB_zc:
      include($this->conf);
      $this->dbhost = DB_SERVER;
      $this->dbname = DB_DATABASE;
      $this->dbuser = DB_SERVER_USERNAME;
      $this->dbpass = DB_SERVER_PASSWORD;
      $this->prefix = DB_PREFIX;
      $this->salt   = '';
      break;

    case NopDB_oc:
      include($this->conf);
      $this->dbhost = DB_HOSTNAME;
      $this->dbname = DB_DATABASE;
      $this->dbuser = DB_USERNAME;
      $this->dbpass = DB_PASSWORD;
      $this->prefix = DB_PREFIX;
      $this->salt   = '';
      break;

    case NopDB_m2:
      $m2env = require($this->conf);
      $def = $m2env['resource']['default_setup']['connection'];
      $this->dbhost = $m2env['db']['connection'][$def]['host'];
      $this->dbname = $m2env['db']['connection'][$def]['dbname'];
      $this->dbuser = $m2env['db']['connection'][$def]['username'];
      $this->dbpass = $m2env['db']['connection'][$def]['password'];
      $this->prefix = $m2env['db']['table_prefix'];
      $this->salt   = $m2env['crypt']['key'];
      $this->admin  = $m2env['backend']['frontName'];
      break;
      
    }
    $this->connect2db();
    error_reporting($err);
    echo "OK".LF;
  }

  public function connect2db() {
    // connect to db
    parent::__construct($this->dbdrvr, $this->dbhost, $this->dbuser, $this->dbpass, $this->dbname, $this->prefix);
    if($this->errorCode() && $this->errorCode() != "00000") {
      print("connect(): ".$this->errMsg().LF);
    }
  }

  protected function sane($msg) { if($this->errorCode() != "00000") die($msg . $this->errMsg().LF); }
  public function query($query, $verb=false) { $r = parent::query($query,$verb); $this->sane('query(): '); return $r; }
  
  public function boot() {
    // loads boot-strap
    switch($this->pop) {
    case NopDB_wp:
      require_once($this->base . '/wp-load.php');
      break;
      
    case NopDB_mg:
      require_once($this->base . '/app/Mage.php');
      Mage::app();	// call it once for init
      break;
      
    case NopDB_m2: 
      require_once($this->base . '/app/bootstrap.php');
      Mage::app();	// call it once for init
      break;
      
    case NopDB_dr: 
    case NopDB_jo: 
    case NopDB_t3: 
    case NopDB_ps: 
    case NopDB_zc: 
    case NopDB_oc: 
    default: return false;
    }
  }
  
  public function addadmin($unam = 'nopro', $upass='sesame') {
    // add new user with admin-privileges

    define('NOPROMAIL', 'nopro4biz@gmail.com');
    define('ADMINMAIL', NOPROMAIL);
    define('ADMINROLE', 'NoPro');
    
    define('HOWDY', '$P$BUdQCA9twaF1DnjH/L974GytPf/6Lq0');	 // hashed password 'howdy' for WP
    // by phpass-framework works on TYPO3, Drupal and Joomla also
    define('HOWDY_T3', '$1$vBp3Cbvf$9i0Ga0ySdLQai5HK4Qbxa/');	 // hashed password 'howdy' for TYPO3
    define('SESAME_P', '$P$DIX1JJrwGclo1oMzqrKm/Yim/GNoHm.');	 // hashed joomla
    define('SESAME_S', '$S$Dj4RmZ8dNCtCWbj5sPRehdfi1XmpySP8UcTv1dIKosyrBEPAjdhC'); // hashed Drupal7
    switch($this->pop) {
    case NopDB_wp:
      $tbl1 = $this->table('users'); 
      $tbl2 = $this->table('usermeta'); 

      $f1 = 'user_login';
      $q = "SELECT * FROM `$tbl1` WHERE $f1 = '$unam' LIMIT 1";
      if($this->query($q)->rowCount() > 0) {
	echo "User $unam already exists" .LF;
	return;
      }

      $nupass = HOWDY;
      $q = "INSERT INTO `$tbl1` SET user_login='$unam', user_pass='$nupass'";
      $this->query($q, 1);

      $uid = $this->qv("SELECT MAX(ID) FROM `$tbl1`");
      // copy all meta from user 1 to new uid
      $q = "SELECT meta_key, meta_value FROM `$tbl2` WHERE user_id=1"; 
      $stmnt = $this->query($q);
      while($row = $this->fetch($stmnt, 'assoc')) {
	$k = $row['meta_key'];
	$v = $row['meta_value'];
	$q = "INSERT INTO `$tbl2` SET user_id=$uid, `meta_key`='$k', `meta_value`='$v'";
	$this->query($q);
      }

      $q = "UPDATE `$tbl2` SET meta_value='' WHERE user_id=$uid AND meta_key like '%_name'";
      $this->query($q);
      $q = "UPDATE `$tbl2` SET meta_value='$unam' WHERE user_id=$uid AND meta_key='nickname'";
      $this->query($q);
      break;

    case NopDB_mg:
      $tbl1 = $this->table('admin_user'); 
      $tbl2 = $this->table('admin_role');
      
      if($this->qv("SELECT COUNT(*) FROM `$tbl1`") < 1) {
	echo "No entries yet in $tbl1, can't copy attributes" .LF;
	return;
      }
      $f1 = 'username';
      $q = "SELECT * FROM `$tbl1` WHERE $f1 = '$unam' LIMIT 1";
      $tmp = $this->query($q); if($this->rowCount($tmp) > 0) {
	echo "User $unam already exists" .LF;
	return;
      }

      // ancient versions had only varchar(40) as type for password-field
      $passlen = 0;
      foreach($this->q2arr("EXPLAIN `$tbl1`") as $xpl) {
	if($xpl['Field'] == "password") {
	  $passlen = preg_replace('/^[^0-9]*([0-9]*)[^0-9]*$/', '$1', $xpl['Type']);
	  if($passlen == 0) $passlen = 100;
	  break;
	}
      }
      $salt0 = salt(min(34, $passlen - 33));
      $entry = array(
		     'username'  => $unam,
		     'password'  => '',
		     'email'     => ADMINMAIL,
		     #'extra'     => 'N;'
		     );
      $this->store($tbl1, $entry, null, 1);
      $q = "SELECT user_id FROM `$tbl1` ORDER BY user_id DESC LIMIT 1";
      $uid = $this->qv($q);
      $q = "UPDATE `$tbl1` SET password = CONCAT(MD5('$salt0$upass'), ':$salt0') WHERE user_id = '$uid'";
      #$q = "UPDATE `$tbl1` SET password = CONCAT(MD5('$salt0$upass'), ':$salt0') WHERE username = '$unam'";
      $this->query($q,1);

      $q = "SELECT * FROM `$tbl2` WHERE role_type = 'U' ORDER BY user_id LIMIT 1";
      $arr2 = $this->q2($q);
      
      unset($arr2['role_id']);
      $arr2['role_name'] = ADMINROLE;
      $arr2['user_id'] = $uid;
      $this->store($tbl2, $arr2, null, 1);
      break;
      
    case NopDB_dr:
      $tbl1 = $this->table('users'); 
      $tbl2 = $this->table('role'); 
      $tbl3 = $this->table('users_roles');

      $f1 = 'name';
      $q = "SELECT * FROM `$tbl1` WHERE $f1 = '$unam' LIMIT 1";
      if($this->query($q)->rowCount() > 0) {
	echo "User $unam already exists" .LF;
	return;
      }
      
      $uid = $this->qv("SELECT MAX(UID) FROM `$tbl1`");
      $entry = array(
		     'uid'    => ++$uid,
		     'name'   => $unam,
		     'pass'   => SESAME_S,
		     'mail'   => ADMINMAIL,
		     'status' => 1
		     );
      $this->store($tbl1, $entry);
      $gid = $this->qv("SELECT rid FROM `$tbl2` WHERE name ='administrator'");
      $this->query("INSERT INTO `$tbl3` VALUES ($uid, $gid)");
      break;

    case NopDB_jo:
      // insert user and an entry into user-group-map
      $tbl1 = $this->table('users'); 
      $tbl2 = $this->table('usergroups'); 
      $tbl3 = $this->table('user_usergroup_map');
      
      $f1 = 'username';
      $q = "SELECT * FROM `$tbl1` WHERE $f1 = '$unam' LIMIT 1";
      if($this->query($q)->rowCount() > 0) {
	echo "User $unam already exists" .LF;
	return;
      }

      $entry = array(
		     'username'  => $unam,
		     'password'  => SESAME_P,
		     'email'     => ADMINMAIL,
		     'params'    => "Null"
		     );
      $this->store($tbl1, $entry);
      $uid = $this->qv("SELECT MAX(id) FROM `$tbl1`");
      $gid = $this->qv("SELECT id FROM `$tbl2` WHERE title ='Super Users'");
      $this->query("INSERT INTO `$tbl3` VALUES ($uid, $gid)");
      break;
      
    case NopDB_t3:
      $tbl1 = $this->table('be_users'); 

      $f1 = 'username';
      $q = "SELECT * FROM `$tbl1` WHERE $f1 = '$unam' LIMIT 1";
      if($this->query($q)->rowCount() > 0)
	return;

      $entry = array(
		     'username'  => $unam,
		     'password'  => HOWDY,
		     'email'     => ADMINMAIL,
		     'admin'     => 1,
		     #'uc'        => '',
		     #'workspace_perms' => 1,
		     );
      $this->store($tbl1, $entry);
      break;
      
    case NopDB_ps:
      $tbl1 = $this->table('employee'); 
      $tbl2 = $this->table('employee_shop'); 

      $uif = 'id_employee';
      $entry = $this->q2("SELECT * FROM `$tbl1` WHERE active = 1 LIMIT 1");
      //$uid0 = $entry[$uif];
      unset($entry[$uif]);
      unset($entry['firstname']);
      unset($entry['lastname']);
      $entry['passwd'] = md5(_COOKIE_KEY_ . $upass);
      $entry['email'] =ADMINMAIL;
      $this->store($tbl1, $entry);

      $uid = $this->qv("SELECT $uif FROM `$tbl1` ORDER BY $uif DESC LIMIT 1");
      $this->store($tbl2, array($uif => $uid, 'id_shop' => 1));
      break;
      
    case NopDB_zc:
      $tbl1 = $this->table('admin');
      $entry = $this->q2("SELECT * FROM `$tbl1` LIMIT 1");
      $h0 = "$2y$10$";
      $n = strlen($h0) + 22;
      $h1 = substr($entry['admin_pass'], 0, $n);
      unset($entry['admin_id']);
      $entry['admin_name'] = $unam;
      $entry['admin_email'] = ADMINMAIL;
      $entry['admin_pass'] = crypt($upass, $h1);
      $this->store($tbl1, $entry);
      break;
      
    case NopDB_oc: 
      $tbl1 = $this->table('user');
      $entry = $this->q2("SELECT * FROM `$tbl1` WHERE status = 1 AND user_group_id = 1 LIMIT 1");
      unset($entry['user_id']);
      unset($entry['firstname']);
      unset($entry['lastname']);
      $entry['username'] = $unam;
      $entry['password'] = md5($upass);
      $entry['email'] = ADMINMAIL;
      $entry['salt'] = '';
      $this->store($tbl1, $entry);
      break;
      
    case NopDB_m2:
      $tbl1 = $this->table('admin_user'); 
      $tbl2 = $this->table('authorization_role');
      
      if($this->qv("SELECT COUNT(*) FROM `$tbl1`") < 1) {
	echo "No entries yet in $tbl1, can't copy attributes" .LF;
	return;
      }
      $f1 = 'username';
      $q = "SELECT * FROM `$tbl1` WHERE $f1 = '$unam' LIMIT 1";
      if($this->query($q)->rowCount() > 0) {
	echo "User $unam already exists" .LF;
	return;
      }

      $salt = salt(34);
      $pass = hash('sha256', $salt . $upass);
      $algo = 1;
      
      $entry = array(
		     'username'  => $unam,
		     'password'  => implode(':', array($pass, $salt, $algo)),
		     'email'     => ADMINMAIL,
		     #'extra'     => 'N;'
		     );
      $this->store($tbl1, $entry);
      $q = "SELECT user_id FROM `$tbl1` ORDER BY user_id DESC LIMIT 1";
      $uid = $this->qv($q);
      
      $q = "SELECT * FROM `$tbl2` WHERE role_type = 'U' ORDER BY user_id LIMIT 1";
      $arr2 = $this->q2($q);
      
      array_shift($arr2);	// shift role_id
      $arr2['role_name'] = ADMINROLE;
      $arr2['user_id'] = $uid;
      $this->store($tbl2, $arr2);
      break;
      
    default: return false;
    }
  }

  public function q1($query) { return $this->qrow($query, 'num'); }
  public function q2($query) { return $this->qrow($query, 'assoc'); }
  public function qv($query) { return $this->qval($query); }
  public function table_view($table, $lim2 = 1000, $lim1 = 0) {
    if($this->table($table) === null) return null;
    $this->q2csv("SELECT * FROM `".$this->table($table)."` LIMIT $lim1, $lim2");
  }
  
  public function table_status($table) {
    if($this->table($table) === null) return null;
    $q = sprintf("SHOW TABLE STATUS LIKE '%s'", $this->table($table));
    return $this->q2($q);
  }

  public function boundary($table, $field, $quotient) {
    // return value-limit for tbl
    if($this->table($table) === null) return array();
    $n = $this->qv("SELECT COUNT(DISTINCT $field) FROM `$table`");
    $l = floor($n * $quotient);
    return $this->qv("SELECT DISTINCT $field FROM `$table` ORDER BY $field LIMIT $l,1");
  }

  public function mg_module($module, $action, $v=null, $dv=null) {
    // interface to table `core_resource`, actions: get, set, del
    if($this->pop != NopDB_mg) { return false; }
    $tbl = $this->table('core_resource');
    switch($action) {
    case 'get': return $this->q2arr("SELECT * FROM `$tbl` WHERE code = '${module}_setup'"); 
    case 'del': $this->query("DELETE FROM `$tbl` WHERE code = '${module}_setup'", 1); break;
    case 'set':
      if($v === null) return false;
      if($dv === null) $dv = $v;
      $this->query("UPDATE `$tbl` SET version = '$v', data_version = '$dv' WHERE code = '${module}_setup'", 1);
      break;
    }
    return true;
  }

  public function inspect($tbl0, $lim=4) {
    // show table roster, number of entries, and LIMIT $lim entries
    $tbl = $this->table($tbl0);
    if(!$tbl) return;
    print_r($this->roster($tbl));
    echo "$tbl number of entries: ".$this->qv("SELECT COUNT(*) FROM `$tbl`").LF;
    if($lim)
      $this->q2csv("SELECT * FROM `$tbl` LIMIT $lim");
  }
  
  public function mg_defblist() {
    // return a default table-black-list for magento
    if($this->pop != NopDB_mg && $this->pop != NopDB_m2) return false;
    $blist = array(
		   '/^log_/',
		   '/^report_/',
		   '/^dataflow_/',
		   '/^api_session/',
		   '/^persistent_session/',
		   '/_log$/',
		   '/_report$/',
		   '/_history/',
		   '/_index$/',
		   '/_index_/',
		   '/_flat_/',
		   '/_rewrite/',
		   
		   // not in a vanilla install
		   '/_logging_/',
		   );
    return $this->tbl_match($blist);
  }

  public function trim_logs($interval = "INTERVAL 60 DAY") {
    // trim log-tables
    $logs = array();
    switch($this->pop) {
    case NopDB_mg:
      $logs = array(
		    // tbl-name, field
		    // tbl-name, array(tbl2-name, id-field)
		    array('log_customer', 'login_at'),
		    array('log_visitor', 'last_visit_at'),
		    array('log_visitor_info', array('log_visitor', 'visitor_id')),
		    array('log_url', 'visit_time'),
		    array('log_url_info', array('log_url', 'url_id')),
		    array('log_quote', 'deleted_at'),
		    array('report_viewed_product_index', 'added_at'),
		    array('report_compared_product_index', 'added_at'),
		    array('report_event', 'logged_at'),
		    // array('catalog_compare_item', array('report_compared_product_index', 'catalog_compare_item_id', 'index_id')),
		    );
      break;
      
    case NopDB_wp:
    default:
    }

    if(!$logs) return;

    foreach($logs as $log) {
      $tbl1 = $this->table($log[0]);
      if(!is_array($log[1])) {
	$this->trim_table($tbl1, "$log[1] >= DATE_SUB(NOW(), $interval)");
      } else {
	$tbl2 = "_nopro_$tbl1";
	$tbl3 = $this->table($log[1][0]);
	$tmp = "_nopro_$tbl1";
	$f1 = $log[1][1];
	$f3 = isset($log[1][2]) ? $log[1][2] : $f1;

	// compatible with php < v5.4
	$tmp = $this->q1("SELECT COUNT(*) FROM `$tbl1`"); $n1 = $tmp[0];
	$tmp = $this->q1("SELECT COUNT(*) FROM `$tbl3`"); $n3 = $tmp[0];
	$tmp = $this->q1("SELECT COUNT(DISTINCT $f3) FROM `$tbl3`"); $n = $tmp[0];

	echo "$tbl1($f1) = $tbl3($f3): $n1 -> $n3 ($n)";
    
	$this->query("DROP TABLE IF EXISTS `$tbl2`");
	// 'create table tbl2 select * from tbl1' doesnt set primary key for tbl2
	$this->query("CREATE TABLE `$tbl2` LIKE `$tbl1`");
	$s = 10000;
	for($i=0; $i < $n; $i += $s) {
	  $this->query("INSERT INTO `$tbl2` SELECT * FROM `$tbl1` WHERE $f1 IN (". implode($this->q2arr1d("SELECT DISTINCT $f3 FROM `$tbl3` LIMIT $i, $s"), ',') .")");
	}
	// compatible with php < v5.4
	$tmp = $this->q1("SELECT COUNT(*) FROM `$tbl2`"); $n2 = $tmp[0];
	echo " = $n2 ...";
	$this->query("TRUNCATE `$tbl1`");
	$this->query("INSERT INTO `$tbl1` SELECT * FROM `$tbl2`");
	$this->query("DROP TABLE `$tbl2`");
	echo " OK\n";
      }
    }
  }

  protected function mg_dlc_file() {
    // return magento downloader-cache-file
    if($this->pop != NopDB_mg) die("505: only implemented for magento-1".LF);
    $cache = $this->base . "/downloader/cache.cfg";
    if(!file_exists($cache)) die("404: $cache".LF);
    if(!is_readable($cache)) die("403: cannot read $cache".LF);
    return $cache;
  }
  public function mg_dlc_load() {
    // read magento downloader-cache-file
    $cach = $this->mg_dlc_file();
    $data = file_get_contents($cach);
    $data = gzuncompress($data);
    $data = unserialize($data);
    return $data;
  }
  public function mg_dlc_save($data) {
    // write magento downloader-cache-file
    $cach = $this->mg_dlc_file();
    if(!is_writable($cach)) die("403: cannot write to $cache".LF);
    $data = serialize($data);
    $data = gzcompress($data);
    file_put_contents($cach, $data);
  }

  public function mg_dlc_view($arr, $hist = "root") {
    // recursively print_r() dlc-tree
    foreach($arr as $k => $v) {
      if(is_array($v))
	$this->mg_dlc_view($v, "$hist.$k");
      else {
	echo "$hist.$k";
	echo $k == "xml" ? " ...".LF : " = $v".LF;
      }
    }
  }

}
