<?php

/**
 * NoPro sql-lib
 * 
 * acts as a meta - abstraction-layer
 * no inheritage here, but choosing one db-extension from mysqli / PDO
 * 
 * 2017-02-04 v0.1 based on NopSQLO / NopSQLP
 * provisorial: uses $dbo = NopSQLO / NopSQLP
 * 
 * 
 * --- history of nopro_sqlo.php
 * 2015-10-10 v0.1 based on the procedural nopro_sqli.php
 *
 * --- history of nopro_sqli.php
 * some handy funcs using mysqli-object
 * 2014-06-24 stopped supporting nopro_sql.php, it's mysqli only now
 * 2015-08-22 sqlistore() now handles an array as $idkey
 * 2015-09-20 added sqli2out()
 * ---
 *
 * 2016-05-06 q2arr() now uses mysqli::fetch_all() instead of a loop
 * 2016-05-07 added q2tbl()
 * 2016-05-29 look if mysqli::fetch_all() exists
 * 2016-05-31 added tbl_sizes_q()
 * 2016-06-04 moved gumshoe() from nopro_pops to here
 * 2016-07-19 moved optimize() from nopro_pops to here
 *            added trim_table()
 * 2016-11-24 fix in roster() and gumshoe(): wrap $tbl inside `` in sql-query "explain `$tbl`"
 * 2016-12-15 adopted table() from nopro_pops
 * 2016-12-22 store(): reordered things to identify and skip an empty data-array
 * 2017-01-31 ported from NopDB (nopro_pops.php): some properties, tables(), table() and tbl_match()
 *            introduced property prefix_skips honoring it in tables()
 *            introduced __construct() (with prefix and prefix_skips)
 * 2017-02-01 added _gumback() as a default callback for gumshoe()
 * 2017-02-03 __construct(): added persistent connection and ssl to remote hosts
 * ---
 *
 *
 * 2017-02-04 added meta-interfaces errMsg(), errorCode(), rowCount($r) fetch($r, type)
 *            added fifi()
 * 2017-02-21 added meta-interfaces lastInsertId()
 * 2017-03-31 added type to qrow()
 * 2018-04-06 __construct: extended params (for dbdrvr "mysqli")
 *            added dbdrvr "SOCK" (for dbo-access via unix-socket)
 * 2018-09-12 gumshoe(): introduced param sqlptrn
 */

require_once('nopro_lib.php');
class NopSQL {

  public
    $dbhost  = null,
    $dbname  = null,
    $dbuser  = null,
    $dbpass  = null,
    $dbport  = null,
    $dbsock  = null,
    $prefix = "",
    $prefix_skips = array()
    ;

  public #protected
    $dbdrvr = null,	// db-extension type
    $tbl = array()
    ;

  #protected $dbo = null;
  public $dbo = null;

  /**/
  public function __destruct() {
    if(!$this->dbo) return;
    debug_msg("-- closing DB-connection --".LF);
    if($this->dbdrvr == "mysqli")
    #if($this->ping())
      $this->close();	// desirable for persistent connections
  }
  /**/
  #public function __destruct() { $this->dbo = null; }
  public function __construct($dbdrvr, $dbhost, $dbuser, $dbpass, $dbname = null, $prefix = "", $prefix_skips = array(), $dbport=null, $dbsock=null) {
    // uses persistent connection and ssl to remote hosts
    $this->dbdrvr = $dbdrvr;
    if(!$dbhost) $dbhost = "localhost";
    $this->dbhost = $dbhost;
    $this->dbuser = $dbuser;
    $this->dbpass = $dbpass;
    $this->dbname = $dbname;
    $this->dbport = $dbport;
    $this->dbsock = $dbsock;
    $this->prefix = $prefix;
    $this->prefix_skips = $prefix_skips;

    switch($this->dbdrvr) {

    case "SOCK":
      #$dsn = "mysql:host=localhost;dbname=$dbname;unix_socket=$dbhost";
      $dsn = "mysql:unix_socket=$dbhost";
      if($dbname) $dsn .= ";dbname=$dbname";
      $opts = array();
      $this->dbo = new PDO($dsn, $dbuser, $dbpass, $opts);
      break;

    case "PDO":
      $dsn = "mysql:dbname=$dbname;host=$dbhost";
      $opts = array();
      if($dbhost != "localhost") {
	$opts[PDO::ATTR_PERSISTENT] = true;
	$opts[PDO::MYSQL_ATTR_COMPRESS] = true;
	#$opts[PDO::MYSQL_ATTR_SSL_CIPHER] = "DHE-RSA-AES256-SHA:AES128-SHA";
	echo "-- no SSL-connection --".LF;
      }
      $this->dbo = new PDO($dsn, $dbuser, $dbpass, $opts);
      break;

    case "mysqli":
      if($dbhost == "localhost")
	$this->dbo = new mysqli($this->dbhost, $this->dbuser, $this->dbpass, $this->dbname, $this->dbport, $this->dbsock);
      else {
	$this->dbo = new mysqli();
	$err = error_reporting(E_ALL & ~E_WARNING);
	$flags = MYSQLI_CLIENT_SSL | MYSQLI_CLIENT_COMPRESS;
	//$this->options(MYSQLI_OPT_SSL_VERIFY_SERVER_CERT, 0);
	if(!$this->real_connect("p:".$this->dbhost, $this->dbuser, $this->dbpass, $this->dbname, $this->dbport, $this->dbsock, $flags)) {
	  echo "-- no SSL-connection: ".$this->errMsg().LF;
	  $flags &= ~MYSQLI_CLIENT_SSL;
	  $this->real_connect("p:".$this->dbhost, $this->dbuser, $this->dbpass, $this->dbname, $this->dbport, $this->dbsock, $flags);
	}
	error_reporting($err);
      }
      break;

    default:
      throw new Exception("DB-Ext \"{$this->dbdrvr}\" not defined (use mysqli/PDO)");
    }
  }

  public function __call($method, $a) {
    // magic method which calls existing dbo->methods
    if(!$this->dbo) return; #die("not connected");
      
    if(0):
    echo "method: "; dbg($method);
    echo "args: "; dbg($a);
    echo "args = ".func_num_args().LF;
    endif;
    if(!method_exists($this->dbo, $method)) throw new Exception("call to undefined method \"$method\"");

    // ugly, but efficient
    switch(count($a)) {
    case 0: return $this->dbo->$method(); break;
    case 1: return $this->dbo->$method($a[0]); break;
    case 2: return $this->dbo->$method($a[0], $a[1]); break;
    case 3: return $this->dbo->$method($a[0], $a[1], $a[2]); break;
    case 4: return $this->dbo->$method($a[0], $a[1], $a[2], $a[3]); break;
    case 5: return $this->dbo->$method($a[0], $a[1], $a[2], $a[3], $a[4]); break;
    case 6: return $this->dbo->$method($a[0], $a[1], $a[2], $a[3], $a[4], $a[5]); break;
    case 7: return $this->dbo->$method($a[0], $a[1], $a[2], $a[3], $a[4], $a[5], $a[6]); break;
    case 8: return $this->dbo->$method($a[0], $a[1], $a[2], $a[3], $a[4], $a[5], $a[6], $a[7]); break;
    default:
      echo "$method(): "; dbg($args);
      die("too many arguments");
    }
  }

  // some meta-interfaces to dbo
  public function errorCode() {
    switch($this->dbdrvr) {
    case "PDO": return $this->dbo->errorCode();
    case "mysqli": return $this->dbo->errno;
    }
  }
  public function errMsg() {
    switch($this->dbdrvr) {
    case "mysqli": return $this->dbo->error;
    case "PDO":
      // backward compatible with php < v5.4:
      $tmp = $this->dbo->errorInfo();
      return $tmp[2];
    }
  }
  public function rowCount($r) {
    switch($this->dbdrvr) {
    case "PDO": return $r->rowCount();
    case "mysqli": return $r->num_rows;
    }
  }
  public function fetch($r, $type) {
    switch($this->dbdrvr) {
    case "PDO": return $r->fetch($type == 'num' ? PDO::FETCH_NUM : PDO::FETCH_ASSOC);
    case "mysqli": return $type == 'num' ? $r->fetch_row() : $r->fetch_assoc();
    }
  }
  public function lastInsertId() {
    switch($this->dbdrvr) {
    case "PDO": return $this->dbo->lastInsertId();
    case "mysqli": return $this->dbo->insert_id;
    }
  }

  public function query($q, $verb=false) {
    if($verb)
      echo $q.LF;
    $rv = $this->dbo->query($q);
    if($rv === false) {
      echo $this->errMsg().LF;
    }
    return $rv;
  }

  public function qrow($q, $type = 'num') {
    if($type == 'num') {
      switch($this->dbdrvr) {
      case 'PDO': return $this->query($q)->fetch(PDO::FETCH_NUM); 
      case 'mysqli': return $this->query($q)->fetch_row(); 
      }
    } else {
      switch($this->dbdrvr) {
      case 'PDO': return $this->query($q)->fetch(PDO::FETCH_ASSOC); 
      case 'mysqli': return $this->query($q)->fetch_assoc(); 
      }
    }
  }
  public function qval($q) {
    // backward compatible with php < v5.4:
    $tmp = $this->qrow($q);
    return $tmp[0]; 
  }

  public function q2csv($q, $out = '/dev/stdout') {
    // output query-result as csv
    $stmnt = $this->query($q) OR die("query(): ".$this->errMsg().LF);
    if(!is_bool($stmnt) && $this->rowCount($stmnt) > 0) {
      $fh = fopen($out, 'w') OR $fh = STDOUT; #die("ERR: fopen($out).LF");
      $row = $this->fetch($stmnt, 'assoc');
      fputcsv($fh, array_keys($row));
      fputcsv($fh, array_values($row));
      while($row = $this->fetch($stmnt, 'num')) {
	fputcsv($fh, $row);
      }
      fclose($fh);
    }
  }
  
  public function q2out($query) {
    // echo sql-results
    $stmnt = $this->query($query) OR die("query(): ".$this->errMsg().LF);
    $row = $this->fetch($stmnt, 'assoc');
    if($row)
      foreach(array_keys($row) as $k)
	echo "$k\t";
    echo LF;
    while($row) {
      foreach($row as $v)
	echo "$v\t";
      echo LF;
      $row = $this->fetch($stmnt, 'num');
    }
  }
  
  public function store($tbl, $arr, $idkey = null, $verbose = false, $keys = null) {
    /* the presence of $idkey, signals UPDATE instead of INSERT
     * $verbose means to print the query
     * $keys for non-associative $arr, or to override $k from $arr
     */
    if(!$idkey)
      $cond = "";
    else {
      if(is_array($idkey)) {
	$cond = 'WHERE';
	foreach($idkey as $k) {
	  $cond .= " `$k` = '$arr[$k]' AND";
	  unset($arr[$k]);
	}
	$cond = substr($cond, 0, strlen($query) - 4);
      } else {
	$cond = sprintf(" WHERE `$idkey` = '%s'", $arr[$idkey]);
	unset($arr[$idkey]);
      }
    }
    if(!$arr) {
      if($verbose)
	echo "No data to store".LF;
      return;
    }

    $query = $idkey ? "UPDATE `$tbl` SET" : "INSERT INTO `$tbl` SET";
    foreach($arr as $k => $val) {
      $key = $keys ? $keys[$k] : $k;
      if($key == $idkey || (is_array($idkey) && in_array($key, $idkey)))
	continue;
      if(is_numeric($val) || $val == "NOW()")
	$query .= " `$key` = $val,";
      else if($val === null)
	$query .= " `$key` = NULL,";
      else
	$query .= " `$key` = '$val',";
    }
    $query = rtrim($query, ',') . $cond;
    
    if($verbose)
      echo $query.LF;
    #die($query.LF);
    
    $this->query($query) OR die("store(): ". $this->errMsg().LF);
    #return ($idkey && !is_array($idkey)) ? $arr[$idkey] : $this->dbo->insert_id;
  }

  public function q2arr($query, $assoc=true) {
    // fill array with sql-results
    $stmnt = $this->query($query) OR die("query(): ".$this->errMsg().LF);
    switch($this->dbdrvr) {
    case 'PDO': return $stmnt->fetchAll($assoc ? PDO::FETCH_ASSOC : PDO::FETCH_NUM);
    case 'mysqli':
    if(method_exists($stmnt, 'fetch_all'))
      return $stmnt->fetch_all($assoc ? MYSQLI_ASSOC : MYSQLI_NUM);
    // loop over results
    $arr = array();
    if($assoc)
      while($row = $stmnt->fetch_assoc()) { array_push($arr, $row); }
    else
      while($row = $stmnt->fetch_row()) { array_push($arr, $row); }
    return $arr;
    }
  }

  public function q2arr1d($query) {
    // fill 1d-array with sql-results
    $arr = array();
    $stmnt = $this->query($query) OR die("query(): ".$this->errMsg().LF);
    while($row = $this->fetch($stmnt, 'num')) {
      array_push($arr, $row[0]);
    }
    return $arr;
  }

  public function tables($upd = false, $all = false) {
    // load tbl[] with table_names matching prefix, honoring prefix_skips
    if($upd) $this->tbl = array();
    if(!$this->tbl) {
      if(!$all && ($this->prefix || $this->prefix_skips)) {
	foreach($this->q2arr1d('SHOW TABLES') as $tbl) {
	  if($this->prefix_skips) {
	    foreach($this->prefix_skips as $prefix)
	      if(strpos($tbl, $prefix) === 0)
		continue 2;
	  }
	  if($this->prefix && strpos($tbl, $this->prefix) !== 0)
	    continue;
	  $this->tbl[] = $tbl;
	}
      } else
	$this->tbl = $this->q2arr1d('SHOW TABLES');
    }
    return $this->tbl;
  }

  public function table($table) {
    if(in_array($table, $this->tables()))
      return $table;
    else if(in_array($this->prefix . $table, $this->tables()))
      return $this->prefix . $table;
    else
      return null;
  }

  public function tbl_match($regexes, $arr = array(), $pref=true) {  // $arr _NOT_ by ref
    // return list of tables matching regex
    if($pref)
      $regexes = preg_replace('/\^/', '^'.$this->prefix, $regexes);
    foreach($this->tables() as $tbl) {
      foreach($regexes as $regex)
	if(preg_match($regex, $tbl)) {
	  $arr[] = $tbl;
	  break;
	}
    }
    return array_unique($arr);
  }

  public function roster($tbl, &$prim = false) {
    // return array of fname => ftyype
    $ret = array();
    $xpl = $this->q2arr("EXPLAIN `$tbl`");   // DESCRIBE table
    foreach($xpl as $arr) {
      $ret[$arr['Field']] = $arr['Type'];
    }
    if($prim !== false) $prim = sql_find_pri($xpl);
    return $ret;
  }

  public function coll($query, $dim = 1) {
    // return array ('collection') as result of query
    return $dim > 1 ? $this->q2arr($query) : $this->q2arr1d($query);
  }

  public function setXchecks($on, $verb=false) {
    // switch foreign key-checks 
    if(!$on)
      $this->query("SET @OLD_FOREIGN_KEY_CHECKS=@@FOREIGN_KEY_CHECKS, FOREIGN_KEY_CHECKS=0", $verb);
    else
      $this->query("SET FOREIGN_KEY_CHECKS=@OLD_FOREIGN_KEY_CHECKS", $verb);
  }

  public function q2tbl($query, $cap='q2tbl()') { arr2tbl($this->q2arr($query), $cap); }

  public function tbl_sizes_q($mb=0) {
    // return query only for info on table-sizes bigger than mb
    // NOT honoring prefix or prefix_skips
    $q = "SELECT ROUND((data_length + index_length) / 1024/1024) AS MB";
    //$q .= ",  sum(MB) as total";
    $q .= ", data_length/1024 AS 'data in KB', index_length/1024 AS 'index in KB'";
    $q .= ", table_name";
    $q .= " FROM information_schema.TABLES WHERE table_schema = '". $this->dbname ."'";
    $q .= " AND data_length > 1024*1024*$mb";
    $q .= " ORDER BY MB, data_length";
    return $q;
  }

  protected function _gumback($data, $retval) {
    // default callback for gumshoe()
    unset($data->nopdb); // we dont wanna output this
    if($data->num > 1) unset($data->cols, $data->key, $data->prim, $data->total);
    echo LF;
    print_r($data);
    return $retval;
  }
  public function gumshoe($nemo, $wlist, $blist, $callback, $sqlptrn = false) {
    // search for nemo in all text-fields of all tables(), limited by wlist and blist
    /**
     * return-values of the function callback($params)
     * 0 continue result-loop
     * 1 break result-loop, continue search-loop
     * >1 end search-loop, continue table-loop
     * <0 return from gumshoe()
     */
    if(!$nemo) return false;
    foreach($this->tables() as $tbl) {
      // table-loop
      if($wlist && !in_array($tbl, $wlist)) continue;
      if($blist && in_array($tbl, $blist)) continue;
      $xpl = $this->q2arr("EXPLAIN `$tbl`");
      $prim = sql_find_pri($xpl);
      if(!$prim) {
	echo "No primary idx in $tbl" .LF;
	continue;
      }
      
      fprintf(STDERR, "%-70s".CR, $tbl);
      $look = sql_find_txtfields($xpl);
      $total = $this->qval("SELECT COUNT(*) FROM `$tbl`");
      foreach($nemo as $key => $ptrn) {
	// search-loop
	if(!$sqlptrn)
	  $ptrn = preg_replace('/([_%])/', '\\\$1', $ptrn);  // escape '%' and '_'
	$cond = 'WHERE';
	foreach($look as $col) {
	  if(!$sqlptrn)
	    $cond .= " `$col` LIKE '%$ptrn%' OR";
	  else
	    $cond .= " `$col` LIKE '$ptrn' OR";
	}
	$cond .= " FALSE";
	$stmnt = $this->query("SELECT * FROM `$tbl` $cond");
	if($this->rowCount($stmnt) > 0) echo $tbl;
	$rv = 0;
	for($i=1; $row = $this->fetch($stmnt, 'assoc'); $i++) {
	  // result-loop
	  if($callback === null) {
	    echo " [{$row[$prim]}]";
	    continue;
	  }
	  $data = (object) array(
				 'key'   => $key,                // idx of pattern-array
				 'ptrn'  => $ptrn,               // val of pattern-array
				 'nopdb' => $this,               // back-link
				 'table' => $tbl,                // db-table
				 'prim'  => $prim,               // primary key in table
				 'num'   => $i,                  // num of current match
				 'found' => $this->rowCount($stmnt), // count of all matches
				 'total' => $total,              // count of tbl-entries
				 'data'  => $row,                // table-row
				 'cols'  => $look);              // names of text-fields in table
	  $rv = is_numeric($callback) ? $this->_gumback($data, $callback) : call_user_func($callback, $data);
	  if($rv) break; // end result-loop
	}
	if($this->rowCount($stmnt) > 0) echo LF;
	if($rv < 0) return; // bye bye
	if($rv > 1) break;  // end search-loop
      }
    }
  }

  public function optimize($verb = false) {
    // cleanup mysql-tables
    foreach($this->tables() as $tbl) {
      $this->query("OPTIMIZE TABLE `$tbl`", $verb);
    }
  }

  public function trim_table($tbl, $cond, $verb=true) {
    /* for instance keep only entries younger than 60 days:

CREATE TABLE table2 AS SELECT * FROM table1 WHERE updateTime >= NOW() - INTERVAL 60 DAY;	-- DATE_SUB(NOW(), $interval)
TRUNCATE table1;
INSERT INTO table1 SELECT * FROM table2;
DROP TABLE table2;

    */
    
    // 'create table tbl2 select * from tbl1' doesnt set primary key for tbl2
    // but it doesnt matter here, cause its a temp-table only

    $tbl1 = $this->table($tbl);
    $tbl2 = "_nopro_$tbl1";

    if($verb) echo "trim $tbl1: ";
    $this->query("DROP TABLE IF EXISTS `$tbl2`");
    // this only works if sql-mode NO_ZERO_DATE is _not_ set
    $this->query("CREATE TABLE `$tbl2` AS SELECT * FROM `$tbl1` WHERE $cond");
    // compatible with php < v5.4
    $tmp = $this->q2arr1d("SELECT COUNT(*) FROM `$tbl1`"); $n1 = $tmp[0];
    $tmp = $this->q2arr1d("SELECT COUNT(*) FROM `$tbl2`"); $n2 = $tmp[0];

    $this->query("TRUNCATE `$tbl1`");
    $this->query("INSERT INTO `$tbl1` SELECT * FROM `$tbl2`");
    $this->query("DROP TABLE `$tbl2`");
    
    if($verb) echo "$n1 rows -> $n2 rows".LF;
  }

  public function fifi($regex, $set = array()) {
    // findfield: search for field-name in subset of tables()
    if(!$set) $set = $this->tables();
    $found = array();
    foreach($set as $tbl) {
      fprintf(STDERR, "%-70s".CR, $tbl);
      $xpl = $this->q2arr("explain `$tbl`");
      foreach($xpl as $field) {
	if(preg_match($regex, $field['Field']))
	  $found[] = $tbl .".". $field['Field'];
      }
    }
    return $found;
  }
  
}

class NopSQLP extends NopSQL {
  public function __construct($dbhost, $dbuser, $dbpass, $dbname = null, $prefix = "", $prefix_skips = array()) {
    parent::__construct("PDO", $dbhost, $dbuser, $dbpass, $dbname, $prefix, $prefix_skips);
  }  
}

class NopSQLO extends NopSQL {
  public function __construct($dbhost, $dbuser, $dbpass, $dbname = null, $prefix = "", $prefix_skips = array()) {
    parent::__construct("mysqli", $dbhost, $dbuser, $dbpass, $dbname, $prefix, $prefix_skips);
  }  
}
