<?php

/* NoPro sql-lib
 * some handy methods extending mysqli-object
 */

require_once('nopro_lib.php');
// nopro_lib.php must be included here

define("Mysql_SSL_keyPath", "/etc/ssl/mysql/mysql@ffm-ssl");

class NopSQLO extends mysqli {
  
  /**
   * 2015-10-10 v0.1 based on the procedural nopro_sqli.php
   *
   * --- history of nopro_sqli.php
   * some handy funcs using mysqli-object
   * 2014-06-24 stopped supporting nopro_sql.php, it's mysqli only now
   * 2015-08-22 sqlistore() now handles an array as $idkey
   * 2015-09-20 added sqli2out()
   * ---
   *
   * 2016-05-06 q2arr() now uses mysqli::fetch_all() instead of a loop
   * 2016-05-07 added q2tbl()
   * 2016-05-29 look if mysqli::fetch_all() exists
   * 2016-05-31 added tbl_sizes_q()
   * 2016-06-04 moved gumshoe() from nopro_pops to here
   * 2016-07-19 moved optimize() from nopro_pops to here
   *            added trim_table()
   * 2016-11-24 fix in roster() and gumshoe(): wrap $tbl inside `` in sql-query "explain `$tbl`"
   * 2016-12-15 adopted table() from nopro_pops
   * 2016-12-22 store(): reordered things to identify and skip an empty data-array
   * 2017-01-31 ported from NopDB (nopro_pops.php): some properties, tables(), table() and tbl_match()
   *		introduced property prefix_skips honoring it in tables()
   *            introduced __construct() (with prefix and prefix_skips)
   * 2017-02-01 added _gumback() as a default callback for gumshoe()
   * 2017-02-03 __construct(): added persistent connection and ssl to remote hosts
   */

  public
    $dbhost  = null,
    $dbname  = null,
    $dbuser  = null,
    $dbpass  = null,
    $prefix = "",
    $prefix_skips = array()
    ;

  protected
    $tbl = array()
    ;

  public function __destruct() {
    debug_msg("-- closing DB-connection --".LF);
    #if($this->ping())
      $this->close();	// desirable for persistent connections
  }
  public function __construct($dbhost, $dbuser, $dbpass, $dbname = null, $prefix = "", $prefix_skips = array()) {
    // uses persistent connection and ssl to remote hosts
    $this->dbhost = $dbhost;
    $this->dbuser = $dbuser;
    $this->dbpass = $dbpass;
    $this->dbname = $dbname;
    $this->prefix = $prefix;
    $this->prefix_skips = $prefix_skips;
    if(!$dbhost || $dbhost == "localhost")
      parent::__construct($this->dbhost, $this->dbuser, $this->dbpass, $this->dbname);
    else {
      parent::__construct();
      $err = error_reporting(E_ALL & ~E_WARNING);
      $flags = MYSQLI_CLIENT_SSL | MYSQLI_CLIENT_COMPRESS;
      if(!$this->real_connect("p:".$this->dbhost, $this->dbuser, $this->dbpass, $this->dbname, null, null, $flags)) {
	echo "-- no SSL-connection: ".$this->error.LF;
	$flags &= ~MYSQLI_CLIENT_SSL;
	$this->real_connect("p:".$this->dbhost, $this->dbuser, $this->dbpass, $this->dbname, null, null, $flags);
      }
      error_reporting($err);
    }
  }
  
  public function query($q, $verb=false) {
    if($verb)
      echo $q.LF;
    $rv = parent::query($q);
    if($rv === false) {
      echo $this->error.LF;
    }
    return $rv;
  }
  
  public function store($tbl, $arr, $idkey = null, $verbose = false, $keys = null) {
    /* the presence of $idkey, signals UPDATE instead of INSERT
     * $verbose means to print the query
     * $keys for non-associative $arr, or to override $k from $arr
     */
    if(!$idkey)
      $cond = "";
    else {
      if(is_array($idkey)) {
	$cond = 'WHERE';
	foreach($idkey as $k) {
	  $cond .= " `$k` = '$arr[$k]' AND";
	  unset($arr[$k]);
	}
	$cond = substr($cond, 0, strlen($query) - 4);
      } else {
	$cond = sprintf(" WHERE `$idkey` = '%s'", $arr[$idkey]);
	unset($arr[$idkey]);
      }
    }
    if(!$arr) {
      if($verbose)
	echo "No data to store".LF;
      return;
    }

    $query = $idkey ? "UPDATE `$tbl` SET" : "INSERT INTO `$tbl` SET";
    foreach($arr as $k => $val) {
      $key = $keys ? $keys[$k] : $k;
      if($key == $idkey || (is_array($idkey) && in_array($key, $idkey)))
	continue;
      if(is_numeric($val) || $val == "NOW()")
	$query .= " `$key` = $val,";
      else if($val === null)
	$query .= " `$key` = NULL,";
      else
	$query .= " `$key` = '$val',";
    }
    $query = rtrim($query, ',') . $cond;
    
    if($verbose)
      echo $query.LF;
    #die($query.LF);
    
    $this->query($query) OR die("store(): ". $this->error.LF);
    #return ($idkey && !is_array($idkey)) ? $arr[$idkey] : $this->insert_id;
  }
  
  public function q2arr($query, $assoc=true) {
    // fill array with sql-results
    $res = $this->query($query) OR die("query(): ".$this->error.LF);
    if(method_exists($res, 'fetch_all')): // use fetch_all

    $arr = $res->fetch_all($assoc ? MYSQLI_ASSOC : MYSQLI_NUM); 

    else: // use loop

    $arr = array();
    if($assoc)
      while($row = $res->fetch_assoc()) { array_push($arr, $row); }
    else
      while($row = $res->fetch_row()) { array_push($arr, $row); }
    
    endif; // use loop
    
    $res->close();
    return $arr;
  }
  
  public function q2arr1d($query) {
    // fill 1d-array with sql-results
    $arr = array();
    $res = $this->query($query) OR die("query(): ".$this->error.LF);
    while($row = $res->fetch_row()) {
      array_push($arr, $row[0]);
    }
    $res->close();
    return $arr;
  }

  public function tables($upd = false, $all = false) {
    // load tbl[] with table_names matching prefix, honoring prefix_skips
    if($upd) $this->tbl = array();
    if(!$this->tbl) {
      if(!$all && ($this->prefix || $this->prefix_skips)) {
	foreach($this->q2arr1d('SHOW TABLES') as $tbl) {
	  if($this->prefix_skips) {
	    foreach($this->prefix_skips as $prefix)
	      if(strpos($tbl, $prefix) === 0)
		continue 2;
	  }
	  if($this->prefix && strpos($tbl, $this->prefix) !== 0)
	    continue;
	  $this->tbl[] = $tbl;
	}
      } else
	$this->tbl = $this->q2arr1d('SHOW TABLES');
    }
    return $this->tbl;
  }

  public function coll($query, $dim = 1) {
    // return array ('collection') as result of query
    return $dim > 1 ? $this->q2arr($query) : $this->q2arr1d($query);
  }
  
  public function roster($tbl, &$prim = false) {
    // return array of fname => ftyype
    $ret = array();
    $xpl = $this->q2arr("EXPLAIN `$tbl`");   // DESCRIBE table
    foreach($xpl as $arr) {
      $ret[$arr['Field']] = $arr['Type'];
    }
    if($prim !== false) $prim = sql_find_pri($xpl);
    return $ret;
  }
  
  public function q2csv($q, $out = '/dev/stdout') {
    // output query-result as csv
    $res = $this->query($q) OR die("query(): ".$this->error.LF);
    if(!is_bool($res) && $res->num_rows > 0) {
      $fh = fopen($out, 'w') OR $fh = STDOUT; #die("ERR: fopen($out).LF");
      $row = $res->fetch_assoc();
      fputcsv($fh, array_keys($row));
      fputcsv($fh, array_values($row));
      while($row = $res->fetch_row()) {
	fputcsv($fh, $row);
      }
      fclose($fh);
      $res->close();
    }
  }
  
  public function q2out($query) {
    // echo sql-results
    $res = $this->query($query) OR die("query(): ".$this->error.LF);
    $row = $res->fetch_assoc();
    if($row)
      foreach(array_keys($row) as $k)
	echo "$k\t";
    echo LF;
    while($row) {
      foreach($row as $v)
	echo "$v\t";
      echo LF;
      $row = $res->fetch_row();
    }
    $res->close();
  }
  
  public function setXchecks($on, $verb=false) {
    // switch foreign key-checks 
    if(!$on)
      $this->query("SET @OLD_FOREIGN_KEY_CHECKS=@@FOREIGN_KEY_CHECKS, FOREIGN_KEY_CHECKS=0", $verb);
    else
      $this->query("SET FOREIGN_KEY_CHECKS=@OLD_FOREIGN_KEY_CHECKS", $verb);
  }

  public function q2tbl($query, $cap='q2tbl()') { arr2tbl($this->q2arr($query), $cap); }

  public function tbl_sizes_q($mb=0) {
    // return query only for info on table-sizes bigger than mb
    // NOT honoring prefix or prefix_skips
    $q = "SELECT ROUND((data_length + index_length) / 1024/1024) AS MB";
    //$q .= ",  sum(MB) as total";
    $q .= ", data_length/1024 AS 'data in KB', index_length/1024 AS 'index in KB'";
    $q .= ", table_name";
    $q .= " FROM information_schema.TABLES WHERE table_schema = '". $this->dbname ."'";
    $q .= " AND data_length > 1024*1024*$mb";
    $q .= " ORDER BY MB, data_length";
    return $q;
  }

  protected function _gumback($data, $retval) {
    // default callback for gumshoe()
    unset($data->nopdb); // we dont wanna output this
    if($data->num > 1) unset($data->cols, $data->key, $data->prim, $data->total);
    echo LF;
    print_r($data);
    return $retval;
  }
  public function gumshoe($nemo, $wlist, $blist, $callback) {
    // search for nemo in all text-fields of all tables(), limited by wlist and blist
    /**
     * return-values of the function callback($params)
     * 0 continue result-loop
     * 1 break result-loop, continue search-loop
     * >1 end search-loop, continue table-loop
     * <0 return from gumshoe()
     */
    if(!$nemo) return false;
    foreach($this->tables() as $tbl) {
      // table-loop
      if($wlist && !in_array($tbl, $wlist)) continue;
      if($blist && in_array($tbl, $blist)) continue;
      $xpl = $this->q2arr("EXPLAIN `$tbl`");
      $prim = sql_find_pri($xpl);
      if(!$prim) {
	echo "No primary idx in $tbl" .LF;
	continue;
      }
      
      #printf("%-40s".CR, $tbl);
      fwrite(STDERR, sprintf("%-70s".CR, $tbl));
      $look = sql_find_txtfields($xpl);
      //$total = $this->query("SELECT COUNT(*) FROM `$tbl`")->fetch_row()[0];
      // backward compatible with php < v5.4:
      $tmp = $this->query("SELECT COUNT(*) FROM `$tbl`")->fetch_row(); $total = $tmp[0]; 
      
      //
      foreach($nemo as $key => $ptrn) {
	// search-loop
	$ptrn = preg_replace('/([_%])/', '\\\$1', $ptrn);  // escape '%' and '_'
	$cond = 'WHERE';
	foreach($look as $col) {
	  $cond .= " `$col` LIKE '%$ptrn%' OR";
	}
	$cond .= " FALSE";
	$res = $this->query("SELECT * FROM `$tbl` $cond");
	if($res->num_rows > 0) echo $tbl;
	$rv = 0;
	for($i=1; $row = $res->fetch_assoc(); $i++) {
	  // result-loop
	  if($callback === null) {
	    echo " [{$row[$prim]}]";
	    continue;
	  }
	  $data = (object) array(
				 'key'   => $key,                // idx of pattern-array
				 'ptrn'  => $ptrn,               // val of pattern-array
				 'nopdb' => $this,               // back-link
				 'table' => $tbl,                // db-table
				 'prim'  => $prim,               // primary key in table
				 'num'   => $i,                  // num of current match
				 'found' => $res->num_rows,      // count of all matches
				 'total' => $total,              // count of tbl-entries
				 'data'  => $row,                // table-row
				 'cols'  => $look);              // names of text-fields in table
	  $rv = is_numeric($callback) ? $this->_gumback($data, $callback) : call_user_func($callback, $data);
	  if($rv) break; // end result-loop
	}
	if($res->num_rows > 0) echo LF;
	$res->close();
	if($rv < 0) return; // bye bye
	if($rv > 1) break;  // end search-loop
      }
    }
  }

  public function optimize($verb = false) {
    // cleanup mysql-tables
    foreach($this->tables() as $tbl) {
      $this->query("OPTIMIZE TABLE `$tbl`", $verb);
    }
  }

  public function table($table) {
    if(in_array($table, $this->tables()))
      return $table;
    else if(in_array($this->prefix . $table, $this->tables()))
      return $this->prefix . $table;
    else
      return null;
  }
  
  public function trim_table($tbl, $cond, $verb=true) {
    /* for instance keep only entries younger than 60 days:

CREATE TABLE table2 AS SELECT * FROM table1 WHERE updateTime >= NOW() - INTERVAL 60 DAY;	-- DATE_SUB(NOW(), $interval)
TRUNCATE table1;
INSERT INTO table1 SELECT * FROM table2;
DROP TABLE table2;

    */

    $tbl1 = $this->table($tbl);
    $tbl2 = "_nopro_$tbl1";

    if($verb) echo "trim $tbl1: ";
    $this->query("DROP TABLE IF EXISTS $tbl2");
    // this only works if sql-mode NO_ZERO_DATE is _not_ set
    $this->query("CREATE TABLE $tbl2 AS SELECT * FROM $tbl1 WHERE $cond");
    // compatible with php < v5.4
    $tmp = $this->q2arr1d("SELECT COUNT(*) FROM ". $tbl1); $n1 = $tmp[0];
    $tmp = $this->q2arr1d("SELECT COUNT(*) FROM ". $tbl2); $n2 = $tmp[0];

    $this->query("TRUNCATE $tbl1");
    $this->query("INSERT INTO $tbl1 SELECT * FROM $tbl2");
    $this->query("DROP TABLE $tbl2");
    
    if($verb) echo "$n1 rows -> $n2 rows".LF;
  }
  
  public function tbl_match($regexes, $arr = array(), $pref=true) {  // $arr _NOT_ by ref
    // return list of tables matching regex
    if($pref)
      $regexes = preg_replace('/\^/', '^'.$this->prefix, $regexes);
    foreach($this->tables() as $tbl) {
      foreach($regexes as $regex)
	if(preg_match($regex, $tbl)) {
	  $arr[] = $tbl;
	  break;
	}
    }
    return array_unique($arr);
  }

}
