<?php
# list of proxies

/**
 * 2017-01-03 added checkProxy()
 *            getProxy() and torProxy() return reachable proxy only
 * 2017-01-06 store checked proxies in $safe
 * 2017-01-11 introduced $torisup
 * 2017-01-12 introduced feed, addProxy(), remove() and update()
 */

class np_proxy {
  public
    $EOpubs;

  protected
    $proxies,
    $failed,
    $safe,
    $torisup = false,
    $EOprots;

  private
    $EOprivs;

  const feed = "https://proxy-list.org/english/search.php?search=anonymous-and-elite.ssl-yes&country=any&type=anonymous-and-elite&port=any&ssl=yes";

  public function proxies()  { return $this->proxies; }
  public function unreachable()  { return $this->failed; }
  public function checked()  { return $this->safe; }
  
  public function checkProxy($proxy, $verb=1) {
    $rep = error_reporting(0);
    #$sock = stream_socket_client("tcp://$proxy", $errno, $errstr, 15,  STREAM_CLIENT_ASYNC_CONNECT);
    $sock = stream_socket_client("tcp://$proxy", $errno, $errstr, 15);
    error_reporting($rep);
    if(!$sock) {
      if($verb) print("checkProxy($proxy) ERROR: $errno - $errstr\n");
      return false;
    }
    fclose($sock);
    return true;
  }

  public function getProxy() {
    $n = count($this->proxies());
    if($n < 2) die("no proxy reachable\n");
    $x = mt_rand(1, $n-1);
    $proxy = $this->proxies()[$x];
    if(empty($this->safe[$x]) && !$this->checkProxy($proxy, 0)) {
      unset($this->proxies[$x]);
      $this->failed[] = $proxy;
      return $this->getProxy();
    }
    return $this->safe[$x] = $proxy;
  }
  public function torProxy() {
    $proxy = "localhost:9150";
    if(!$this->torisup && !$this->checkProxy($proxy, 1)) die("no proxy\n");
    $this->torisup = true;
    return $proxy;
  }
  
  public function addProxy($proxy) { $this->proxies[] = $proxy; return count($this->proxies); }
  public function remove($nr) { unset($this->proxies[$nr]); return count($this->proxies); }
  
  public function update($refresh = 360) { // default 360 minutes
    require_once('nopro_wget.php');
    list($hdr, $cnt) = npwg_cache($this::feed, null, $refresh);
    for($p1 = $p2 = 0;; $p1 = $p2+1) {
      $p1 = strpos($cnt, "Proxy('", $p1);
      if($p1 === false) break;
      $p2 = strpos($cnt, "')", $p1);
      if($p2 === false) break;
      $s = substr($cnt, $p1+7, $p2-$p1-7);
      $proxy = base64_decode($s);
      if($this->checkProxy($proxy, 0)) {
	$n = $this->addProxy($proxy);
	$this->safe[$n-1] = $proxy;
      }
    }
  }
  
  public function __construct() {
    // http://freeproxylists.net/
    $this->failed = array();
    $this->safe = array();
    $this->proxies = array(''
			   #,'119.188.115.27:80'     // CHN https 1.412 sec 160101 (cloud003IP _spy.php)
			   #,'62.23.15.92:3128'      // FRA https 0.108 sec 160217
			   #,'51.254.103.206:3128'   // GBR https 0.121 sec 160217 // blocked by http://www.naturalworldeco-shop.com/admin,  down 160415
			   /**
			   ,'108.59.10.129:55555'   // USA 0.448 160427
			   ,'108.59.10.141:55555'   // USA 0.459 160427
			   #,'190.54.112.159:3128'   // CHL 1.009 160427 // blocks shoplift: HTTP/1.0 417 Expectation failed Server: squid/2.7.STABLE9
			   ,'61.174.13.12:443'      // CHN 1.243 160428
			   ,'165.139.149.169:3128'  // CHN 0.579 160428
			   #,'119.93.82.148:80'      // PHP 1.104 160428  // blocks shoplift: HTTP/1.0 417 Expectation failed Server: squid/2.7.STABLE9
			   ,'217.196.160.83:3128'   // UKR 0.188 160428
			   ,'117.131.216.214:80'    // CHN 1.240 160428
			   ,'117.131.216.214:8080'  // CHN 1.246 160428
			   /**/
			   #,'77.81.232.85:80'        // ROM 0.798 160518, dead 160526
			   #,'31.173.74.73:8080'      // ROM 0.676, dead 160627
			   #,'149.202.249.227:3128'   // GER 0.224
			   #,'43.226.162.110:80'      // JAP 1.634, 16.596 160607
			   #,'62.117.96.138:3128'     // RUS 0.899, dead 160607
			   #,'91.134.143.169:3128'    // BUL 0.078
			   #,'81.88.142.24:8080'      // SLO 0.124, dead 160607
			   #,'61.8.35.166:8088'       // AUS 2.3
			   #,'158.85.241.82:1080'     // USA 0.399, dead 160607
			   #,'223.67.136.218:80'      // CHI 1.241, dead 160623
			   #,'81.82.240.15:8080'      // BEL 0.431, dead 160627
			   #,'46.101.129.227:3128'    // RUS 0.042, dead 160623
			   #,'189.28.166.79:80'       // BRA 1.450, 404 160623
			   #,'178.32.220.216:3128'    // FRA 0.370 160623, dead 160627
			   #,'43.226.162.106:8080'    // JAP 1.183, dead 160710
			   #,'202.148.13.2:80'        // IDN, dead 160710
			   #,'81.196.2.213:3128'      // ROM 0.178, dead 160710
			   #,'67.52.216.91:3128'      // USA 0.703, auth required 160627
			   #,'200.151.249.242:80'     // BRA 1.246, dead 160710
			   #,'81.196.2.213:3128'      // ROM 0.301, dead 160710
			   
			   #,'186.170.31.134:8080'    // COL 1.174 160607, dead 160722
			   #,'201.55.46.6:80'         // BRA 1.200 160627, dead 160725
			   #,'205.202.253.128:8080'   // USA 0.827, requires auth 160715
			   /* permission denied 160723
			   ,'202.47.236.250:8080'    // THA 0.906 160715
			   ,'202.47.236.251:8080'    // THA 0.906 160715
			   ,'202.47.236.252:8080'    // THA 0.906 160715
			   */
			   #,'139.59.8.222:666'       // AUS 0.740 160715, dead 160717
			   #,'187.33.231.206:80'      // RUS 1.210 160715, dead 160717
			   #,'47.88.104.219:80'        // CAN 0.731 160717, dead 160905
			   #,'213.108.33.121:3128'     // RUS 0.284 160717, dead 160719
			   #,'210.186.135.229:8080'    // MAL 1.673 160723, dead 160725
			   #,'195.214.249.244:8888'    // BUL 1.184 160723, dead 160802
			   #,'108.59.10.129:55555'     // USA 0.460
			   #,'117.169.17.242:80'       // CHN 1.206, dead 160905
			   #,'88.157.149.250:8080'     // POR 0.617 160802
			   #,'47.88.104.219:80'        // CAN 0.724, dead 160905
			   
			   #,'201.55.46.6:80'          // BRA 1.250 
			   #,'138.201.117.135:3128'    // GER 0.124, dead 160905
			   #,'109.63.189.187:3128'     // RUS 2.216 160905, dead 161031
			   #,'91.205.173.222:3128'     // DEU 161031
			   #,'197.97.146.62:8080'      // ZAF 0.803 161102, dead 170102
			   #,'92.222.70.213:3128'      // DEU 0.278 161114, dead 161208
			   #,'91.205.173.222:3128'     // DEU 0.732 161114, dead 161208
			   #,'193.219.103.113:8080'    // ESP 0.214 161114, dead 161121
			   #,'193.219.103.114:8080'    // ESP 0.228 161114
			   #,'193.219.103.115:8080'    // ESP 0.194 161114
			   #,'193.219.103.116:8080'    // ESP 0.211 161114
			   #,'120.92.3.127:80'         // CHN 1.200, dead 170103
			   #,'185.28.193.95:8080'      // EU 0.097
			   #,'12.33.254.195:3128'      // USA 0.476 170102, dead 170104
			   #,'187.53.61.108:3182'      // BRA 0.683 170102, dead 170103
			   #,'217.33.216.114:8080'     // UK 1.554 170103, dead 170109
			   # ,'67.97.220.180:10000'     // USA 0.538, dead 170108
			   #,'200.68.27.100:3128'      // CHI 1.526, dead 170104
			   #,'87.244.194.94:8080'      // SLO 1.466 170104, dead 170106
			   #,'94.177.172.141:8080'     // ROM 1.742, dead 170112
			   #,'144.217.162.239:3128'    // USA 0.541 170108, dead 170109
			   /*
			   ,''
			   /**/
			   );
  }
}
